"""Utility functions."""

from cgi import escape
import sys
import warnings

from zope.container.contained import notifyContainerModified
from zope.event import notify
from zope.lifecycleevent import ObjectMovedEvent

from Acquisition import aq_base, aq_inner, aq_parent
from App.Dialogs import MessageDialog
from OFS.event import ObjectWillBeMovedEvent
from OFS.CopySupport import sanity_check, CopyError
from ZODB.POSException import ConflictError


def unrestricted_move(self, ob):
    """Move an object from one container to another bypassing certain
    checks."""
    orig_id = ob.getId()
    if not ob.cb_isMoveable():
        raise CopyError('Not supported {}'.format(escape(orig_id)))

    try:
        ob._notifyOfCopyTo(self, op=1)
    except ConflictError:
        raise
    except:
        raise CopyError(MessageDialog(
            title="Move Error",
            message=sys.exc_info()[1],
            action='manage_main'))

    if not sanity_check(self, ob):
        raise CopyError("This object cannot be pasted into itself")

    orig_container = aq_parent(aq_inner(ob))
    if aq_base(orig_container) is aq_base(self):
        id = orig_id
    else:
        id = self._get_id(orig_id)

    notify(ObjectWillBeMovedEvent(ob, orig_container, orig_id,
                                  self, id))

    # try to make ownership explicit so that it gets carried
    # along to the new location if needed.
    ob.manage_changeOwnershipType(explicit=1)

    try:
        orig_container._delObject(orig_id, suppress_events=True)
    except TypeError:
        # BBB: removed in Zope 2.11
        orig_container._delObject(orig_id)
        warnings.warn(
            "%s._delObject without suppress_events is deprecated "
            "and will be removed in Zope 2.11." %
            orig_container.__class__.__name__, DeprecationWarning)
    ob = aq_base(ob)
    ob._setId(id)

    try:
        self._setObject(id, ob, set_owner=0, suppress_events=True)
    except TypeError:
        # BBB: removed in Zope 2.11
        self._setObject(id, ob, set_owner=0)
        warnings.warn(
            "%s._setObject without suppress_events is deprecated "
            "and will be removed in Zope 2.11." %
            self.__class__.__name__, DeprecationWarning)
    ob = self._getOb(id)

    notify(ObjectMovedEvent(ob, orig_container, orig_id, self, id))
    notifyContainerModified(orig_container)
    if aq_base(orig_container) is not aq_base(self):
        notifyContainerModified(self)

    ob._postCopy(self, op=1)
    # try to make ownership implicit if possible
    ob.manage_changeOwnershipType(explicit=0)


def getSavedAttrName(attrName):
    return '_old_%s' % attrName


def patch(context, originalAttrName, replacement):
    """ Default handler that preserves original method """
    OLD_NAME = getSavedAttrName(originalAttrName)
    if not hasattr(context, OLD_NAME):
        setattr(context, OLD_NAME, getattr(context, originalAttrName))
    setattr(context, originalAttrName, replacement)


def undoPatch(context, originalAttrName):
    OLD_NAME = getSavedAttrName(originalAttrName)
    setattr(context, originalAttrName, getattr(context, OLD_NAME))
    delattr(context, OLD_NAME)



from Products.Archetypes.Storage import AttributeStorage

def migrateField(obj, action, newObj=None, **kwargs):
    """Migrate an Archetypes field

    - 'obj' should be the object to migrate from
    - 'newObj', if given, should be the object to migrate to. If not given,
        migration happens within obj only. If 'newObj' is given, attributes
        will not be unset on obj even if a 'newFieldName' for an action is
        specified.
    - 'action' specifies an action to apply to the object. It should be a dict,
        with the keys:

            - fieldName (required)
            - storage (optional; default = AttributeStorage)
            - callBefore
            - transform
            - newFieldName
            - newStorage (optional; default = same as 'storage')
            - unset (optional; default False)
            - callAfter

        For each object found by the query, the migrator will test for the
        existence of a field given by 'fieldName' stored using 'storage', an
        Archetypes storage, which defaults to AttributeStorage.

        If this field is found in the given storage, migration of will take
        place.

        If 'newFieldName' is given, this gives the new field name to store using
        'newStorage', if given. If 'newStorage' is not given, it defaults to
        the same storage as 'storage'.

        If 'newFieldName' is not given but 'newStorage' is, this is equivalent
        to switching the storage on a field.

        If callBefore is given, it is called before migration. This should be
        method with the signature:

            callBefore(obj, fieldName, value, **kwargs)

        If this returns True, migration continues. If it returns False, migration
        of this field is skipped.

        If 'transform' is given, this is used to transform the value of the
        attribute between the two attributes. This should be a method with the
        following signature

            transform(obj, value, **kwargs)

        The kwargs will contain any passed-in kwargs, as well as the fields
        'oldFieldName', 'newObj', 'newFieldName'. Note that 'newObj' and
        'newFieldName' may be the same as 'obj' and 'oldFieldName',
        respectively, if no object- or fieldname-migration is taking place.

        This method should return the transformed value of the field. If
        'transform' isgiven and 'newFieldName' is not, the return value from
        transform() is stored back in 'fieldName'.

        If 'unset' is given and is True, the old field is unset from the old
        object, always. This can be used to unset a field when there is no
        newFieldName. Note that 'unset' overrides 'transform', 'newFieldName',
        and 'newStorage'. If 'unset' is given, the value is unset, full stop.

        Finally, if 'callAfter' is given, it is called with the same arguments
        as callBefore().

    Returns True if the field was migrated, False otherwise.
    """

    fieldName = action['fieldName']
    storage = action.get('storage', AttributeStorage())
    newStorage = action.get('newStorage', None)
    callBefore = action.get('callBefore', None)
    transform = action.get('transform', None)
    newFieldName = action.get('newFieldName', None)
    callAfter = action.get('callAfter', None)
    unsetField = action.get('unset', False)

    if newObj is None or obj.getPhysicalPath() == newObj.getPhysicalPath():
        newObj = obj

    if newFieldName is None:
        newFieldName = fieldName

    if newStorage is None:
        newStorage = storage

    doTransform = True
    if transform is None:
        doTransform = False

    try:
        value = storage.get(fieldName, obj)
    except AttributeError:
        return False

    # Call and test callBefore, if given
    if callBefore is not None:
        status = callBefore(obj, fieldName, value, **kwargs)
        if not status:
            return False

    # Unset the old value. We may immediately re-set with the same storage
    # and/or the same fieldName, but this keeps things simple. Also note
    # that we unset the old storage on the possibly new object in case it
    # ended up there during blind attribute migration or initialisation
    storage.unset(fieldName, obj)
    try: storage.unset(fieldName, newObj)
    except (AttributeError, KeyError): pass

    # If we are simply unsetting, skip setting values
    if not unsetField:
        # Apply transform, if given
        if doTransform:
            value = transform(obj, value, fieldName = fieldName,
                                          newObj = newObj,
                                          newFieldName = newFieldName, **kwargs)

        # Now set the possibly transformed value, with the possibly new field name
        # and/or storage on the possibly new object, unless 'unset' is given.
        newStorage.set(newFieldName, newObj, value)

    # Call callAfter, if given
    if callAfter is not None:
        callAfter(newObj, fieldName, value, **kwargs)

    return True



from Products.ATContentTypes.interfaces import IATDocument
from Products.Archetypes.public import StringField
from Products.Archetypes.public import StringWidget
from archetypes.schemaextender.field import ExtensionField
from archetypes.schemaextender.interfaces import ISchemaExtender
from zope.component import adapts
from zope.interface import implementer


class MyStringField(ExtensionField, StringField):
    """A trivial field."""


@implementer(ISchemaExtender)
class PageExtender(object):
    adapts(IATDocument)  # XXX Might not work on Plone 3.

    fields = [
        MyStringField(
            "extensive_text",
            default="default text",
            widget=StringWidget(
                label="This page has extensive text")),
            ]

    def __init__(self, context):
        self.context = context

    def getFields(self):
        return self.fields



# initialize cmf test content
from Products.CMFCore import permissions

def initialize(context):
    import Products.contentmigration.testcontent
    from Products.CMFCore import utils
    utils.ContentInit(
            'CMF Default Content',
            content_types=(),
            permission=permissions.AddPortalContent,
            extra_constructors=(testcontent.addDocument,),
            visibility=None,
            ).initialize(context)



"""Migration tools for ATContentTypes

Migration system for the migration from CMFDefault/Event types to archetypes
based CMFPloneTypes (http://plone.org/products/atcontenttypes/).

Copyright (c) 2004-2005, Christian Heimes <tiran@cheimes.de> and contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.
"""
__author__ = 'Christian Heimes <tiran@cheimes.de>'
__docformat__ = 'restructuredtext'

from copy import copy
import logging

from Acquisition import aq_base
from Acquisition import aq_parent
from Acquisition import aq_inner
from DateTime import DateTime
from Persistence import PersistentMapping
from zope.component import queryAdapter
from zope.component import queryUtility
from OFS.Uninstalled import BrokenClass
from OFS.interfaces import IOrderedContainer
from ZODB.POSException import ConflictError
from zExceptions import BadRequest
from AccessControl.Permission import Permission
from AccessControl import SpecialUsers
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.CMFCatalogAware import WorkflowAware
from Products.contentmigration.common import unrestricted_rename
from Products.contentmigration.common import _createObjectByType
from Products.contentmigration.utils import patch, undoPatch
from Products.Archetypes.interfaces import IReferenceable
from plone.locking.interfaces import ILockable
from plone.uuid.interfaces import IMutableUUID
try:
    from plone.app.redirector.interfaces import IRedirectionStorage
    IRedirectionStorage  # pyflakes
except ImportError:
    IRedirectionStorage = None
try:
    from Products.Archetypes.config import UUID_ATTR
except ImportError:
    UUID_ATTR = None

LOG = logging.getLogger('ATCT.migration')

_marker = []

# Dublin Core mapping
# accessor, mutator, fieldname
DC_MAPPING = (
    ('Title', 'setTitle', None),
    ('Creator', None, None),
    ('Subject', 'setSubject', 'subject'),
    ('Description', 'setDescription', 'description'),
    ('Publisher', None, None),
    ('Contributors', 'setContributors', 'contributors'),
    ('Date', None, None),
    ('CreationDate', None, None),
    ('EffectiveDate', 'setEffectiveDate', 'effectiveDate'),
    ('ExpirationDate', 'setExpirationDate', 'expirationDate'),
    ('ModificationDate', None, None),
    ('Type', None, None),
    ('Format', 'setFormat', None),
    ('Identifier', None, None),
    ('Language', 'setLanguage', 'language'),
    ('Rights', 'setRights', 'rights'),

    # allowDiscussion is not part of the official DC metadata set
    #('allowDiscussion','isDiscussable','allowDiscussion'),
  )

# Mapping used from DC migration
METADATA_MAPPING = []
for accessor, mutator, field in DC_MAPPING:
    if accessor and mutator:
        METADATA_MAPPING.append((accessor, mutator))


def copyPermMap(old):
    """bullet proof copy
    """
    new = PersistentMapping()
    for k, v in old.items():
        new[k] = v
    return new


def getPermissionMapping(acperm):
    """Converts the list from ac_inherited_permissions into a dict
    """
    result = {}
    for entry in acperm:
        result[entry[0]] = entry[1]
    return result


class BaseMigrator(object):
    """Migrates an object to the new type

    The base class has the following attributes:

     * src_portal_type
       The portal type name the migration is migrating *from*
       Can be overwritten in the constructor
     * src_meta_type
       The meta type of the src object
     * dst_portal_type
       The portal type name the migration is migrating *to*
       Can be overwritten in the constructor
     * dst_meta_type
       The meta type of the dst object
     * map
       A dict which maps old attributes/function names to new
     * old
       The old object which has to be migrated
     * new
       The new object created by the migration
     * parent
       The folder where the old objects lives
     * orig_id
       The id of the old objet before migration
     * old_id
       The id of the old object while migrating
     * new_id
       The id of the new object while and after the migration
     * kwargs
       A dict of additional keyword arguments applied to the migrator
    """
    src_portal_type = None
    src_meta_type = None
    dst_portal_type = None
    dst_meta_type = None
    map = {}

    def __init__(self, obj, src_portal_type=None, dst_portal_type=None,
                 **kwargs):
        self.old = aq_inner(obj)
        self.orig_id = self.old.getId()
        self.old_id = '%s_MIGRATION_' % self.orig_id
        self.new = None
        self.new_id = self.orig_id
        self.parent = aq_parent(self.old)
        if src_portal_type is not None:
            self.src_portal_type = src_portal_type
        if dst_portal_type is not None:
            self.dst_portal_type = dst_portal_type
        self.kwargs = kwargs

        # safe id generation
        while hasattr(aq_base(self.parent), self.old_id):
            self.old_id += 'X'
        msg = "%s (%s -> %s)" % (self.old.absolute_url(1), self.src_portal_type,
                                 self.dst_portal_type)
        LOG.debug(msg)

    def getMigrationMethods(self):
        """Calculates a nested list of callables used to migrate the old object
        """
        beforeChange = []
        methods = []
        lastmethods = []
        for name in dir(self):
            if name.startswith('beforeChange_'):
                method = getattr(self, name)
                if callable(method):
                    beforeChange.append(method)
            if name.startswith('migrate_'):
                method = getattr(self, name)
                if callable(method):
                    methods.append(method)
            if name.startswith('last_migrate_'):
                method = getattr(self, name)
                if callable(method):
                    lastmethods.append(method)

        afterChange = methods + [self.custom, self.finalize] + lastmethods
        return (beforeChange, afterChange, )

    def migrate(self, unittest=0):
        """Migrates the object
        """
        beforeChange, afterChange = self.getMigrationMethods()

        # Unlock according to plone.locking:
        lockable = ILockable(self.old, None)
        if lockable and lockable.locked():
            lockable.unlock()
        # Unlock according to webdav:
        if self.old.wl_isLocked():
            self.old.wl_clearLocks()

        for method in beforeChange:
            __traceback_info__ = (self, method, self.old, self.orig_id)
            # may raise an exception, catch it later
            method()
        # preserve position on rename
        self.need_order = IOrderedContainer.providedBy(self.parent)
        if self.need_order:
            self._position = self.parent.getObjectPosition(self.orig_id)
        self.renameOld()
        self.createNew()

        for method in afterChange:
            __traceback_info__ = (self, method, self.old, self.orig_id)
            # may raise an exception, catch it later
            method()

        self.reorder()
        self.remove()

    __call__ = migrate

    def renameOld(self):
        """Renames the old object

        Must be implemented by the real Migrator
        """
        raise NotImplementedError

    def createNew(self):
        """Create the new object

        Must be implemented by the real Migrator
        """
        raise NotImplementedError

    def custom(self):
        """For custom migration
        """
        pass

    def migrate_properties(self):
        """Migrates zope properties

        Removes the old (if exists) and adds a new
        """
        if not hasattr(aq_base(self.old), 'propertyIds') or \
          not hasattr(aq_base(self.new), '_delProperty'):
            # no properties available
            return None

        for id in self.old.propertyIds():
            if id in ('title', 'description', 'content_type', ):
                # migrated by dc or other
                continue
            value = self.old.getProperty(id)
            typ = self.old.getPropertyType(id)
            __traceback_info__ = (self.new, id, value, typ)
            if self.new.hasProperty(id):
                self.new._delProperty(id)
            # continue if the object already has this attribute
            if getattr(aq_base(self.new), id, _marker) is not _marker:
                continue
            try:
                self.new._setProperty(id, value, typ)
            except ConflictError:
                raise
            except:
                LOG.error('Failed to set property %s type %s to %s at object %s' %
                          (id, typ, value, self.new), exc_info=True)

    def migrate_owner(self):
        """Migrates the zope owner
        """
        # getWrappedOwner is not always available
        if hasattr(aq_base(self.old), 'getWrappedOwner'):
            owner = self.old.getWrappedOwner()
            self.new.changeOwnership(owner)
        else:
            # fallback
            # not very nice but at least it works
            # trying to get/set the owner via getOwner(), changeOwnership(...)
            # did not work, at least not with plone 1.x, at 1.0.1, zope 2.6.2
            self.new._owner = self.old.getOwner(info=1)

    def migrate_localroles(self):
        """Migrate local roles
        """
        self.new.__ac_local_roles__ = None
        # Clean the auto-generated creators by Archetypes
        # ExtensibleMetadata.  Do set the original creators.
        self.new.setCreators(self.old.listCreators())
        local_roles = self.old.__ac_local_roles__
        if not local_roles:
            # Only set owner local role if the owner can be retrieved
            owner = self.old.getWrappedOwner()
            if owner is not SpecialUsers.nobody:
                self.new.manage_setLocalRoles(owner.getId(), ['Owner'])
        else:
            self.new.__ac_local_roles__ = copy(local_roles)
            if hasattr(self.old, '__ac_local_roles_block__'):
                self.new.__ac_local_roles_block__ = (
                    self.old.__ac_local_roles_block__)

    def migrate_user_roles(self):
        """Migrate roles added by users
        """
        if getattr(aq_base(self.old), 'userdefined_roles', False):
            ac_roles = self.old.userdefined_roles()
            new_ac_roles = tuple(getattr(self.new, '__ac_roles__', ()))
            if ac_roles:
                for role in ac_roles:
                    if role not in new_ac_roles:
                        self.new._addRole(role)

    def migrate_permission_settings(self):
        """Migrate permission settings (permission <-> role)
        The acquire flag is coded into the type of the sequence. If roles is a list
        than the roles are also acquire. If roles is a tuple the roles aren't
        acquired.
        """
        oldmap = getPermissionMapping(self.old.ac_inherited_permissions(1))
        newmap = getPermissionMapping(self.new.ac_inherited_permissions(1))
        for key, values in oldmap.items():
            old_p = Permission(key, values, self.old)
            old_roles = old_p.getRoles()
            new_values = newmap.get(key, ())
            new_p = Permission(key, new_values, self.new)
            new_p.setRoles(old_roles)

    def migrate_withmap(self):
        """Migrates other attributes from obj.__dict__ using a map

        The map can contain both attribute names and method names

        'oldattr' : 'newattr'
            new.newattr = oldattr
        'oldattr' : ''
            new.oldattr = oldattr
        'oldmethod' : 'newattr'
            new.newattr = oldmethod()
        'oldattr' : 'newmethod'
            new.newmethod(oldatt)
        'oldmethod' : 'newmethod'
            new.newmethod(oldmethod())
        """
        for oldKey, newKey in self.map.items():
            #LOG("oldKey: " + str(oldKey) + ", newKey: " + str(newKey))
            if not newKey:
                newKey = oldKey
            oldVal = getattr(self.old, oldKey)
            newVal = getattr(self.new, newKey)
            if callable(oldVal):
                value = oldVal()
            else:
                value = oldVal
            if callable(newVal):
                newVal(value)
            else:
                setattr(self.new, newKey, value)

    def remove(self):
        """Removes the old item

        Must be implemented by the real Migrator
        """
        raise NotImplementedError

    def reorder(self):
        """Reorder the new object in its parent

        Must be implemented by the real Migrator
        """
        raise NotImplementedError

    def finalize(self):
        """Finalize construction (called between)
        """
        ob = self.new
        fti = self.new.getTypeInfo()
        # This calls notifyWorkflowCreated which resets the migrated workflow
        # fti._finishConstruction(self.new)

        # _setPortalTypeName is required to
        if hasattr(ob, '_setPortalTypeName'):
            ob._setPortalTypeName(fti.getId())

        #self.new.reindexObject(['meta_type', 'portal_type'], update_metadata=False)


class BaseCMFMigrator(BaseMigrator):
    """Base migrator for CMF objects
    """

    def migrate_dc(self):
        """Migrates dublin core metadata
           This needs to be done after custom migrations, as you cannot
           setContentType on Images and Files until there is an object stored.
        """
        for accessor, mutator in METADATA_MAPPING:
            oldAcc = getattr(self.old, accessor)
            newMut = getattr(self.new, mutator)
            oldValue = oldAcc()
            newMut(oldValue)

    def migrate_workflow(self):
        """migrate the workflow state
        """
        wfh = getattr(self.old, 'workflow_history', None)
        if wfh:
            wfh = copyPermMap(wfh)
            self.new.workflow_history = wfh

    def migrate_allowDiscussion(self):
        """migrate allow discussion bit
        """
        if getattr(aq_base(self.old), 'allowDiscussion', _marker) is not _marker and \
          getattr(aq_base(self.new), 'isDiscussable', _marker)  is not _marker:
            self.new.isDiscussable(self.old.allowDiscussion())

    def migrate_discussion(self):
        """migrate talkback discussion bit
        """
        talkback = getattr(self.old.aq_inner.aq_explicit, 'talkback', _marker)
        if talkback is _marker:
            return
        else:
            self.new.talkback = talkback

    def beforeChange_storeDates(self):
        """Save creation date and modification date
        """
        self.old_creation_date = self.old.CreationDate()
        self.old_mod_date = self.old.ModificationDate()

    def last_migrate_date(self):
        """migrate creation / last modified date

        Must be called as *last* migration
        """
        self.new.creation_date = DateTime(self.old_creation_date)
        self.new.setModificationDate(DateTime(self.old_mod_date))

    def beforeChange_redirects(self):
        """Load redirects."""
        if IRedirectionStorage is None:
            return
        storage = queryUtility(IRedirectionStorage)
        if storage is None:
            return
        path = '/'.join(self.old.getPhysicalPath())
        self.old_redirects = storage.redirects(path)

    def migrate_redirects(self):
        """migrate redirects
        """
        if IRedirectionStorage is None:
            return
        storage = queryUtility(IRedirectionStorage)
        if storage is None:
            return
        path = '/'.join(self.new.getPhysicalPath())
        for redirect in self.old_redirects:
            storage.add(redirect, path)


class ItemMigrationMixin(object):
    """Migrates a non folderish object
    """
    isFolderish = False

    def renameOld(self):
        """Renames the old object
        """
        #LOG("renameOld | orig_id: " + str(self.orig_id) + "; old_id: " + str(self.old_id))
        #LOG(str(self.old.absolute_url(1)))
        unrestricted_rename(self.parent, self.orig_id, self.old_id)
        #self.parent.manage_renameObject(self.orig_id, self.old_id)

    def createNew(self):
        """Create the new object
        """
        _createObjectByType(self.dst_portal_type, self.parent, self.new_id)
        self.new = getattr(aq_inner(self.parent).aq_explicit, self.new_id)

    def remove(self):
        """Removes the old item
        """
        self.parent.manage_delObjects([self.old_id])

    def reorder(self):
        """Reorder the new object in its parent
        """
        if self.need_order:
            try:
                self.parent.moveObject(self.new_id, self._position)
            except ConflictError:
                raise
            except:
                LOG.error('Failed to reorder object %s in %s' % (self.new,
                          self.parent), exc_info=True)


class FolderMigrationMixin(ItemMigrationMixin):
    """Migrates a folderish object
    """
    isFolderish = True

    def beforeChange_patchNotifyWorkflowCreatedMethod(self):
        def notifyWorkflowCreated(self):
            """
              Do not notifyWorkflowCreated...
            """
            pass

        patch(WorkflowAware, 'notifyWorkflowCreated', notifyWorkflowCreated)

    def beforeChange_storeSubojects(self):
        """store subobjects from old folder
        This methods gets all subojects from the old folder and removes them from the
        old. It also preservers the folder order in a dict.
        For performance reasons the objects are removed from the old folder before it
        is renamed. Elsewise the objects would be reindex more often.
        """

        orderAble = IOrderedContainer.providedBy(self.old)
        orderMap = {}
        subobjs = {}

        # using objectIds() should be safe with BrokenObjects
        for id in self.old.objectIds():
            obj = getattr(self.old.aq_inner.aq_explicit, id)
            # Broken object support. Maybe we are able to migrate them?
            if isinstance(obj, BrokenClass):
                LOG.warning('BrokenObject in %s' % self.old.absolute_url(1))
                #continue

            if orderAble:
                try:
                    orderMap[id] = self.old.getObjectPosition(id) or 0
                except AttributeError:
                    LOG.debug("Broken OrderSupport", exc_info=True)
                    orderAble = 0
            subobjs[id] = aq_base(obj)
            # delOb doesn't call manage_afterAdd which safes some time because it
            # doesn't unindex an object. The migrate children method uses
            # _setObject later. This methods indexes the object again and
            # so updates all catalogs.
        for id in self.old.objectIds():
            # Loop again to remove objects, order is not preserved when
            # deleting objects
            self.old._delOb(id)
            # We need to take care to remove the relevant ids from _objects
            # otherwise objectValues breaks.
            if getattr(self.old, '_objects', None) is not None:
                self.old._objects = tuple([o for o in self.old._objects
                                           if o['id'] != id])

        self.orderMap = orderMap
        self.subobjs = subobjs
        self.orderAble = orderAble

    def migrate_children(self):
        """Copy childish objects from the old folder to the new one
        """
        subobjs = self.subobjs
        for id, obj in subobjs.items():
            # we have to use _setObject instead of _setOb because it adds the object
            # to folder._objects but also reindexes all objects.
            __traceback_info__ = __traceback_info__ = ('migrate_children',
                          self.old, self.orig_id, 'Migrating subobject %s' % id)
            try:
                self.new._setObject(id, obj, set_owner=0)
            except BadRequest:
                # If we already have the object we need to remove it carefully
                # and retry.  We can assume that such an object must be
                # autogenerated by a prior migration step and thus is less
                # correct than the original.  Such objects may want to
                # consider setting the attribute '__replaceable__' to avoid
                # this.
                if id in self.new.objectIds():
                    self.new._delOb(id)
                    if getattr(self.new, '_objects', None) is not None:
                        self.new._objects = tuple([o for o in
                                        self.new._objects if o['id'] != id])
                    self.new._setObject(id, obj, set_owner=0)
                else:
                    raise

        # reorder items
        # in CMF 1.5 Topic is orderable while ATCT's Topic is not orderable
        # order objects only when old *and* new are orderable we can't check
        # when creating the map because self.new == None.
        if self.orderAble and IOrderedContainer.providedBy(self.new):
            orderMap = self.orderMap
            for id, pos in orderMap.items():
                self.new.moveObjectToPosition(id, pos)

    def last_migrate_restoreNotifyWorkflowCreatedMethod(self):
        undoPatch(WorkflowAware, 'notifyWorkflowCreated')


class UIDMigrator(object):
    """Migrator class for migration CMF and AT uids
    """

    def migrate_cmf_uid(self):
        """Migrate CMF uids
        """
        uidhandler = getToolByName(self.parent, 'portal_uidhandler', None)
        if uidhandler is None:
            return  # no uid handler available
        uid = uidhandler.queryUid(self.old, default=None)
        if uid is not None:
            uidhandler.setUid(self.new, uid, check_uniqueness=False)

    def migrate_at_uuid(self):
        """Migrate AT universal uid
        """
        if not IReferenceable.providedBy(self.old):
            return  # old object doesn't support AT uuids
        uid = self.old.UID()
        self.old._uncatalogUID(self.parent)
        if UUID_ATTR:  # Prevent object deletion triggering UID related magic
            setattr(self.old, UUID_ATTR, None)
        if queryAdapter(self.new, IMutableUUID):
            IMutableUUID(self.new).set(str(uid))
        else:
            self.new._setUID(uid)


class CMFItemMigrator(ItemMigrationMixin, UIDMigrator, BaseCMFMigrator):
    """Migrator for items implementing the CMF API
    """


class CMFFolderMigrator(FolderMigrationMixin, UIDMigrator, BaseCMFMigrator):
    """Migrator from folderish items implementing the CMF API
    """



from zope.component import queryUtility
from zope.interface import Interface
try:
    from plone.app.redirector.interfaces import IRedirectionStorage
    IRedirectionStorage  # pyflakes
except ImportError:
    IRedirectionStorage = None


class MarkerInterface(Interface):
    pass


def get_redirects(new_object):
    """Get p.a.redirector redirects.

    Maybe this is also used outside of Plone, or p.a.redirector is
    optional, so try not to fail too hard then.

    The tests will currently fail in that case.  If wanted, we could
    arrange something here, for example letting add_redirects return
    the a list of successfully stored redirects, which would be an
    empty list when p.a.redirector is not there, and letting
    get_redirects return an empty list.
    """
    if IRedirectionStorage is None:
        return
    storage = queryUtility(IRedirectionStorage)
    if storage is None:
        return
    path = '/'.join(new_object.getPhysicalPath())
    return storage.redirects(path)


def add_redirects(old_paths, new_object):
    """Add p.a.redirector redirects.

    Maybe this is also used outside of Plone, or p.a.redirector is
    optional, so try not to fail too hard then.
    """
    if IRedirectionStorage is None:
        return
    storage = queryUtility(IRedirectionStorage)
    if storage is None:
        return
    path = '/'.join(new_object.getPhysicalPath())
    for redirect in old_paths:
        storage.add(redirect, path)



import logging

from Products.ZCatalog.Catalog import mergeResults

from Products.contentmigration.common import HAS_LINGUA_PLONE
from Products.contentmigration.basemigrator.walker import CatalogWalker, registerWalker

LOG = logging.getLogger('contentmigration')


class CustomQueryWalker(CatalogWalker):
    """Walker using portal_catalog and an optional custom query. The ATCT
    migration framework uses this to find content to migrate.
    """

    def __init__(self, portal, migrator, src_portal_type=None, dst_portal_type=None,
                    query={}, callBefore=None, **kwargs):
        """Set up the walker. See contentmigration.basemigrator.walker for details.

        The 'query' parameter can be used to pass a dict with custom catalog
        query parameters. Note that portal_type and meta_type will be set
        based on src_portal_type or, if given, the src_portal_type in the
        migrator.

        The 'callBefore' parameter can be used to pass a function that will
        be called before each item is migrated. If it returns False, the item
        will be skipped. It should have the signature:

            callBefore(oldObject, **kwargs)

        The kwargs passed to this constructor will be passed along to the
        test function.
        """
        CatalogWalker.__init__(self, portal, migrator,
                               src_portal_type=src_portal_type,
                               dst_portal_type=dst_portal_type, **kwargs)
        self.additionalQuery = query
        self.callBefore = callBefore
        self.kwargs = kwargs


    def walk(self):
        """Walks around and returns all objects which needs migration

        :return: objects (with acquisition wrapper) that needs migration
        :rtype: generator
        """
        catalog = self.catalog
        query = self.additionalQuery.copy()
        query['portal_type'] = self.src_portal_type
        query['meta_type'] = self.src_meta_type

        if HAS_LINGUA_PLONE and 'Language' in catalog.indexes():
            query['Language'] = 'all'

        brains = catalog(query)
        limit = getattr(self, 'limit', False)
        if limit:
            brains = brains[:limit]

        # unpack to be able to log brains that break during migration
        brains = [i for i in brains]

        for brain in brains:
            try:
                obj = brain.getObject()
            except AttributeError:
                LOG.error("Couldn't access %s" % brain.getPath())
                continue
            except KeyError:
                LOG.error("Couldn't access RID %s" % brain.getRID())
                continue

            if self.callBefore is not None and callable(self.callBefore):
                if self.callBefore(obj, **self.kwargs) == False:
                    continue

            try:
                state = obj._p_changed
            except:
                state = 0
            if obj is not None:
                yield obj
                # safe my butt
                if state is None:
                    obj._p_deactivate()

registerWalker(CustomQueryWalker)

class MultiCustomQueryWalker(CustomQueryWalker):
    """A catalog walker that combines the results from multiple
    queries."""

    additionalQueries = ()

    def walk(self):
        """Walks around and returns all objects which needs migration

        :return: objects (with acquisition wrapper) that needs migration
        :rtype: generator
        """
        catalog = self.catalog
        query = self.additionalQuery
        query['portal_type'] = self.src_portal_type
        query['meta_type'] = self.src_meta_type

        if HAS_LINGUA_PLONE and 'Language' in catalog.indexes():
            #query['Language'] = catalog.uniqueValuesFor('Language')
            query['Language'] = 'all'

        results = []
        for addQ in self.additionalQueries:
            addQ.update(query)
            results.append(catalog(addQ))
        brains = mergeResults(results, has_sort_keys=False,
                              reverse=False)

        for brain in brains:
            obj = brain.getObject()

            if self.callBefore is not None and callable(self.callBefore):
                if self.callBefore(obj, **self.kwargs) == False:
                    continue

            try: state = obj._p_changed
            except: state = 0
            if obj is not None:
                yield obj
                # safe my butt
                if state is None: obj._p_deactivate()

registerWalker(MultiCustomQueryWalker)



"""Support for migrators that place the destination object in a
different location from the source object."""

import logging
from Acquisition import aq_inner
from ZODB.POSException import ConflictError

from Products.contentmigration.common import _createObjectByType

from Products.contentmigration.inplace import (InplaceCMFItemMigrator,
                                               InplaceCMFFolderMigrator)
logger = logging.getLogger(__name__)

class TranslocatingMigratorMixin(object):
    """A migrator that places the destination object in a different
    location from the source object."""

    def getDestinationParent(self):
        """Return the container into which the destination will be
        added."""
        return self.parent

    def createNew(self):
        """Create the new object
        """
        dst_parent = self.getDestinationParent()

        # Support AT migration if used
        schema = getattr(self, 'schema', {})

        _createObjectByType(self.dst_portal_type, dst_parent,
                            self.new_id, **schema)
        self.new = getattr(aq_inner(dst_parent).aq_explicit, self.new_id)

    def reorder(self):
        """Reorder in the right parent.
        """
        if self.need_order:
            try:
                self.getDestinationParent().moveObject(
                    self.new_id, self._position)
            except ConflictError:
                raise
            except:
                logger.error('Failed to reorder object %s in %s' % (self.new,
                          self.parent), exc_info=True)

class TranslocatingInplaceCMFItemMigrator(TranslocatingMigratorMixin,
                                          InplaceCMFItemMigrator):
    """Inplace migrator for items implementing the CMF API."""
    pass

class TranslocatingInplaceCMFFolderMigrator(
    TranslocatingMigratorMixin, InplaceCMFFolderMigrator):
    """Inplace migrator for folders implementing the CMF API."""
    pass

# BBB: Not sure why these below were done this way, but they are in
# use by remember so I'm leaving them here.

from Products.contentmigration.basemigrator.migrator import UIDMigrator
from Products.contentmigration.inplace import InplaceUIDMigrator

class TranslocatingInplaceMigrator(TranslocatingMigratorMixin,
                                   InplaceUIDMigrator):
    """UID migration support for inplace translocating migrators."""
    pass

class TranslocatingMigrator(TranslocatingMigratorMixin,
                            UIDMigrator):
    """UID migration support for translocating migrators."""
    pass



from Products.contentmigration.tests.cmtc import ContentMigratorTestCase
from Products.contentmigration.tests.cmtc import makeUpper
from Products.contentmigration.tests.cmtc import conditionallyAbortAttribute
from Products.contentmigration.tests.cmtc import conditionallyAbortObject
from Products.contentmigration.tests.cmtc import callAfterAttribute
from Products.Archetypes.Storage.annotation import AnnotationStorage
from Products.Archetypes.Storage import AttributeStorage


class TestFieldMigration(ContentMigratorTestCase):
    """Test migration of in-instance AT fields"""

    def afterSetUp(self):
        # Create some content to migrate
        self.folder.invokeFactory('Document', 'd1')
        self.folder.invokeFactory('Document', 'd2')
        self.folder.invokeFactory('News Item', 'n1')
        self.folder.invokeFactory('News Item', 'n2')

        self.folder['d1'].setTitle('Document 1')
        self.folder['d1'].setDescription('Description 1')
        self.folder['d1'].setText('Body one')

        self.folder['d2'].setTitle('Document 2')
        self.folder['d2'].setDescription('Description 2')
        self.folder['d2'].setText('Body two')

        self.folder['n1'].setTitle('News 1')
        self.folder['n1'].setDescription('Description 3')
        self.folder['n1'].setText('News one')

        self.folder['n2'].setTitle('News 2')
        self.folder['n2'].setDescription('Description 4')
        self.folder['n2'].setText('News two')

        if 'front-page' in self.portal.objectIds():
            self.portal._delObject('front-page')
        self.setRoles(['Manager', 'Member'])

    def testAttributeRenaming(self):
        storage = AnnotationStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'newFieldName' : 'bodyText',
                    'storage'      : storage
                    },)

        self.execute(query, actions)
        self.assertEqual(storage.get('bodyText', self.folder['d1']).getRaw(), 'Body one')
        self.assertEqual(storage.get('bodyText', self.folder['d2']).getRaw(), 'Body two')
        self.assertRaises(AttributeError, storage.get, 'text', self.folder['d1'])

    def testTransform(self):
        storage = AnnotationStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'storage'      : storage,
                    'transform'    : makeUpper
                     },)
        self.execute(query, actions)
        self.assertEqual(storage.get('text', self.folder['d1']).getRaw(), 'BODY ONE')
        self.assertEqual(storage.get('text', self.folder['d2']).getRaw(), 'BODY TWO')

    def testAttributeRenamingAndTransform(self):
        storage = AnnotationStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'newFieldName' : 'bodyText',
                    'storage'      : storage,
                    'transform'    : makeUpper
                     },)
        self.execute(query, actions)
        self.assertEqual(storage.get('bodyText', self.folder['d1']).getRaw(), 'BODY ONE')
        self.assertEqual(storage.get('bodyText', self.folder['d2']).getRaw(), 'BODY TWO')

    def testNewStorageAndAttribute(self):
        storage = AnnotationStorage()
        newStorage = AttributeStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'newFieldName' : 'bodyText',
                    'storage'      : storage,
                    'newStorage'   : newStorage
                    },)
        self.execute(query, actions)
        self.assertEqual(getattr(self.folder['d1'], 'bodyText').getRaw(), 'Body one')
        self.assertEqual(getattr(self.folder['d2'], 'bodyText').getRaw(), 'Body two')
        self.assertRaises(AttributeError, storage.get, 'text', self.folder['d1'])

    def testNewStorageOnly(self):
        storage = AnnotationStorage()
        newStorage = AttributeStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'storage'      : storage,
                    'newStorage'   : newStorage
                    },)
        self.execute(query, actions)
        self.assertEqual(getattr(self.folder['d1'], 'text').getRaw(), 'Body one')
        self.assertEqual(getattr(self.folder['d2'], 'text').getRaw(), 'Body two')

        try:
            storage.get('text', self.folder['d1'])
        except AttributeError:
            pass
        else:
            self.fail()

    def testAbortAttribute(self):
        storage = AnnotationStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'transform'    : makeUpper,
                    'storage'      : storage,
                    'callBefore'   : conditionallyAbortAttribute,
                   },)
        self.execute(query, actions)
        self.assertEqual(storage.get('text', self.folder['d1']).getRaw(), 'Body one')
        self.assertEqual(storage.get('text', self.folder['d2']).getRaw(), 'BODY TWO')

    def testAbortObject(self):
        storage = AnnotationStorage()
        query   = {}
        actions = ({'fieldName'    : 'text',
                    'transform'    : makeUpper,
                     'storage'      : storage,
                    },)
        self.execute(query, actions, conditionallyAbortObject)
        self.assertEqual(storage.get('text', self.folder['d1']).getRaw(), 'Body one')
        self.assertEqual(storage.get('text', self.folder['d2']).getRaw(), 'BODY TWO')


    def testCallAfterAttribute(self):
        storage = AnnotationStorage()
        lst = []
        query = {}
        actions = ({'fieldName'    : 'text',
                    'transform'    : makeUpper,
                    'storage'      : storage,
                    'callAfter'    : callAfterAttribute
                    },)
        self.execute(query, actions, lst = lst)
        lst.sort()
        self.assertEqual(lst, ['d1: text = BODY ONE', 'd2: text = BODY TWO'])

    def testUnsetAttributeWithCustomQuery(self):
        storage = AnnotationStorage()
        query   = {'getId' : 'd1'}
        actions = ({'fieldName'    : 'text',
                    'unset'        : True,
                    'storage'      : storage,
                    },)
        self.execute(query, actions)
        self.assertEqual(storage.get('text', self.folder['d2']).getRaw(), 'Body two')
        with self.assertRaises(AttributeError):
            storage.get('text', self.folder['d1'])



from Acquisition import aq_inner
from field import migrateField


class BaseInlineMigrator(object):
    """This migrator base class is similar to the ATContentTypes BaseMigrator,
    except it is designed to migrate fields within the same content type,
    not across types. To that end, it does not do a rename-copy-create cycle,
    it simply operates on the objects the walker returns.

    The base class has the following attributes:

     * portal_type
       The portal type name the migration is migrating *from*
       Can be overwritten in the constructor
     * meta_type
       The meta type of the src object
     * obj
       The object being migrated
     * kwargs
       A dict of additional keyword arguments applied to the migrator
    """
    portal_type = None
    meta_type = None
    kwargs = {}

    # Framework requires us to have dst_ versions, even if they are not
    # used
    dst_portal_type = dst_meta_type = '__dummy__'

    def __init__(self, obj, portal_type=None, meta_type=None, **kwargs):
        self.obj = aq_inner(obj)
        if portal_type is not None:
            self.portal_type = portal_type
        if meta_type is not None:
            self.meta_type = meta_type
        self.kwargs = kwargs

    def getMigrationMethods(self):
        """Calculates a nested list of callables used to migrate the old object
        """
        beforeChange = []
        methods      = []
        lastmethods  = []
        for name in dir(self):
            if name.startswith('beforeChange_'):
                method = getattr(self, name)
                if callable(method):
                    beforeChange.append(method)
            if name.startswith('migrate_'):
                method = getattr(self, name)
                if callable(method):
                    methods.append(method)
            if name.startswith('last_migrate_'):
                method = getattr(self, name)
                if callable(method):
                    lastmethods.append(method)

        afterChange = methods+[self.custom]+lastmethods
        return (beforeChange, afterChange, )

    def migrate(self):
        """Migrates the object
        """
        beforeChange, afterChange = self.getMigrationMethods()

        for method in beforeChange:
            __traceback_info__ = (self, method, self.obj)
            method()

        for method in afterChange:
            __traceback_info__ = (self, method, self.obj)
            method()

    __call__ = migrate

    def custom(self):
        """Override this to provide custom migration.
        """
        pass

class InlineFieldActionMigrator(BaseInlineMigrator):
    """Mixin class to enable field migrations based on field migration actions.
    This allows migrating of attributes using declarative actions. It works
    on an Archetypes storage level, which means that it's capable of migrating
    attributes when interfaces have changed. It also allows you to migrate
    between different implementations of the same content type, by setting the
    src_- and dest_- portal- and meta- types to the same type.

    To use it, you should set the class variable fieldActions to a list of dicts
    specifying field actions. To see how these work, please look at field.py.

    If you specify field actions, it's probably best not to use the 'map'
    migrator feature for any of these fields.
    """

    fieldActions = ()

    def migrate_withFieldActions(self):
        """Apply field migration actions"""
        for action in self.fieldActions:
            migrateField(obj = self.obj, action = action, **self.kwargs)




import unittest
import doctest

from plone.testing import layered

optionflags = (doctest.NORMALIZE_WHITESPACE |
               doctest.ELLIPSIS |
               doctest.REPORT_ONLY_FIRST_FAILURE |
               doctest.REPORT_NDIFF)

from Products.contentmigration.tests.layer import TestLayer


def test_suite():
    suite = unittest.TestSuite()
    test_files = [
        '../inplace.txt',
        '../translocate.txt',
        '../archetypes.txt',
        '../bugfixes.txt',
        '../schemaextender.txt'
    ]
    for test_file in test_files:
            suite.addTest(layered(
            doctest.DocFileSuite(test_file, package='Products.contentmigration.tests',
                                 optionflags=optionflags),
            layer=TestLayer))
    return suite



from Acquisition import aq_inner
from zope.interface import alsoProvides, directlyProvidedBy

from Products.CMFCore.utils import getToolByName

from Products.Archetypes.ArchetypeTool import getType
from Products.Archetypes.config import REFERENCE_ANNOTATION
from Products.Archetypes.Referenceable import Referenceable

from Products.contentmigration.common import _createObjectByType

from Products.contentmigration.basemigrator.migrator import BaseCMFMigrator
from Products.contentmigration.basemigrator.migrator import ItemMigrationMixin
from Products.contentmigration.basemigrator.migrator import FolderMigrationMixin
from Products.contentmigration.basemigrator.migrator import UIDMigrator

from Products.contentmigration.inplace import BaseInplaceCMFMigrator
from Products.contentmigration.inplace import InplaceItemMigrationMixin
from Products.contentmigration.inplace import InplaceFolderMigrationMixin
from Products.contentmigration.inplace import InplaceUIDMigrator

from Products.contentmigration.translocate import TranslocatingMigratorMixin

_marker = []


def migrate_dummy(self):
    """Do nothing.  Used to override inherited methods we don't
    want to run."""
    pass


class ATMigratorMixin(object):
    """Migrates content of one AT type to another AT type."""

    fields_map = {}
    only_fields_map = False
    accessor_getter = 'getEditAccessor'

    # Override some methods from BaseCMFMigrator that we don't want to
    # run for AT migrations.
    migrate_dc = migrate_dummy
    migrate_allowDiscussion = migrate_dummy

    def beforeChange_storeDates(self):
        """Save creation date and modification date
        """
        old_schema = self.old.Schema()
        for field_name in ['creation_date', 'modification_date']:
            old_field = old_schema[field_name]
            accessor = (old_field.getEditAccessor(self.old) or
                        old_field.getAccessor(self.old))
            setattr(self, 'old_' + field_name, accessor())

    # XXX Can this be set through the call to the constructor?
    def last_migrate_date(self):
        """migrate creation / last modified date

        Must be called as *last* migration
        """
        new_schema = self.new.Schema()
        for field_name in ['creation_date', 'modification_date']:
            new_field = new_schema[field_name]
            mutator = new_field.getMutator(self.new)
            mutator(getattr(self, 'old_' + field_name))

    def beforeChange_references(self):
        """Migrate references annotation."""
        # Set the flag so that references aren't deleted
        self.old._v_cp_refs = 1
        # Move the references annotation storage
        if hasattr(self.old, REFERENCE_ANNOTATION):
            at_references = getattr(self.old, REFERENCE_ANNOTATION)
            setattr(self, REFERENCE_ANNOTATION, at_references)

    def migrate_references(self):
        """Migrate references annotation."""
        # Restor the references annotation
        if hasattr(self, REFERENCE_ANNOTATION):
            at_references = getattr(self, REFERENCE_ANNOTATION)
            setattr(self.new, REFERENCE_ANNOTATION, at_references)
        # Run the reference manage_afterAdd to transition all copied
        # references
        is_cp = getattr(self.old, '_v_is_cp', _marker)
        self.new._v_is_cp = 0
        Referenceable.manage_afterAdd(self.new, self.new,
                                      self.new.__parent__)
        if is_cp is not _marker:
            self.new._v_is_cp = is_cp
        else:
            del self.new._v_is_cp


class BaseATMigrator(ATMigratorMixin, BaseCMFMigrator):
    """Migrates content of one AT type to another AT type."""
    pass


class ATItemMigratorMixin(object):
    """Migrator for items implementing the AT API."""

    def beforeChange_schema(self):
        """Load the values of fields according to fields_map if present.

        Each key in fields_map is a field in the old schema and each
        value is a field in the new schema.  If fields_map isn't a
        mapping, each field in the old schema will be migrated into
        the new schema.  Obeys field modes for readable and writable
        fields.  These values are then passed in as field kwargs into
        the constructor in the createNew method."""

        old_schema = self.old.Schema()

        typesTool = getToolByName(self.parent, 'portal_types')
        fti = typesTool.getTypeInfo(self.dst_portal_type)
        archetype = getType(self.dst_meta_type, fti.product)
        new_schema = archetype['klass'].schema

        if self.only_fields_map:
            old_field_names = self.fields_map.keys()
        else:
            old_field_names = old_schema.keys()

        # Let the migrator handle the id and dates
        for omit_field_name in ['id', 'creation_date',
                                'modification_date']:
            if omit_field_name in old_field_names:
                old_field_names.remove(omit_field_name)

        kwargs = getattr(self, 'schema', {})
        for old_field_name in old_field_names:
            old_field = self.old.getField(old_field_name)
            new_field_name = self.fields_map.get(old_field_name,
                                                 old_field_name)

            if new_field_name is None:
                continue

            new_field = new_schema.get(new_field_name, None)
            if new_field is None:
                continue

            if ('r' in old_field.mode and 'w' in new_field.mode):
                accessor = (
                    getattr(old_field, self.accessor_getter)(self.old)
                    or old_field.getAccessor(self.old))
                value = accessor()
                kwargs[new_field_name] = value
        self.schema = kwargs

    def createNew(self):
        """Create the new object passing in the loaded archetypes
        schema field values as kwargs."""
        _createObjectByType(self.dst_portal_type, self.parent,
                            self.new_id, **self.schema)
        self.new = getattr(aq_inner(self.parent).aq_explicit,
                           self.new_id)

    def migrate_extension_fields(self):
        """Migrate extension fields.

        In the beforeChange_schema method only standard fields of the
        schema are taken care of.  Extension fields from
        archetypes.schemaextender are not handled.  This can only
        really be done after the new object has been created.
        """

        old_schema = self.old.Schema()
        typesTool = getToolByName(self.parent, 'portal_types')
        fti = typesTool.getTypeInfo(self.dst_portal_type)
        archetype = getType(self.dst_meta_type, fti.product)
        plain_new_schema = archetype['klass'].schema
        new_schema = self.new.Schema()

        if self.only_fields_map:
            old_field_names = self.fields_map.keys()
        else:
            old_field_names = old_schema.keys()

        kwargs = {}
        for old_field_name in old_field_names:
            old_field = self.old.getField(old_field_name)
            new_field_name = self.fields_map.get(old_field_name,
                                                 old_field_name)

            if new_field_name is None:
                continue
            if new_field_name in plain_new_schema.keys():
                # Already migrated.
                continue

            new_field = new_schema.get(new_field_name, None)
            if new_field is None:
                continue

            if ('r' in old_field.mode and 'w' in new_field.mode):
                accessor = (
                    getattr(old_field, self.accessor_getter)(self.old)
                    or old_field.getAccessor(self.old))
                value = accessor()
                kwargs[new_field_name] = value
        # Apply the changes.
        self.new.update(**kwargs)

    def migrate_marker_interfaces(self):
        """Migrate marker interfaces.
        """
        alsoProvides(self.new, directlyProvidedBy(self.old))


class ATItemMigrator(ATItemMigratorMixin, ItemMigrationMixin,
                     UIDMigrator, BaseATMigrator):
    """Migrator for items implementing the AT API."""
    pass


class ATFolderMigrator(ATItemMigratorMixin, FolderMigrationMixin, UIDMigrator,
                       BaseATMigrator):
    """Migrator from folderish items implementing the AT API."""
    pass

# Inplace migrators


class BaseInplaceATMigrator(ATMigratorMixin, BaseInplaceCMFMigrator):
    """Migrates content of one AT type to another AT type inplace."""
    pass


class InplaceATItemMigrator(ATItemMigratorMixin,
                            InplaceItemMigrationMixin,
                            InplaceUIDMigrator,
                            BaseInplaceATMigrator):
    """Migrator for items implementing the AT API."""
    pass


class InplaceATFolderMigrator(ATItemMigratorMixin,
                              InplaceFolderMigrationMixin,
                              InplaceUIDMigrator,
                              BaseInplaceATMigrator):
    """Migrator from folderish items implementing the AT API."""
    pass

# Translocating migrators


class TranslocatingInplaceATItemMigrator(TranslocatingMigratorMixin,
                                         InplaceATItemMigrator):
    """Inplace migrator for items implementing the AT API."""
    pass


class TranslocatingInplaceATFolderMigrator(TranslocatingMigratorMixin,
                                           InplaceATFolderMigrator):
    """Inplace migrator for folders implementing the AT API."""
    pass



"""Support for migrators that load all values to be migrated into the
migrator, then delete the old object and then load the values into the
new object without requiring that the old object be renamed or
moved."""

from copy import copy

from Acquisition import aq_base
from ZODB.POSException import ConflictError
from AccessControl.Permission import Permission
from AccessControl import SpecialUsers

from Products.CMFCore.utils import getToolByName

from Products.Archetypes.interfaces import IReferenceable

from Products.contentmigration.basemigrator.migrator import BaseMigrator
from Products.contentmigration.basemigrator.migrator import BaseCMFMigrator
from Products.contentmigration.basemigrator.migrator import ItemMigrationMixin
from Products.contentmigration.basemigrator.migrator import FolderMigrationMixin
from Products.contentmigration.basemigrator.migrator import METADATA_MAPPING
from Products.contentmigration.basemigrator.migrator import getPermissionMapping
from Products.contentmigration.basemigrator.migrator import copyPermMap
from Products.contentmigration.basemigrator.migrator import LOG

_marker = []

class BaseInplaceMigrator(BaseMigrator):
    """Support for migrators that load all values to be migrated into the
    migrator, then delete the old object and then load the values into
    the new object without requiring that the old object be renamed or
    moved."""

    _properties = ()
    _properties_ignored = ('title', 'description', 'content_type', )

    def _checkLoadAttr(self, attr):
        """Check for an attribute on the migrator before loading and
        raise an error if it's already there."""
        if hasattr(self, attr):
            raise(ValueError, ('The migrator already has a value '
                               'for %s.' % attr))

    def beforeChange_properties(self):
        """Load up the migrator with property values."""

        if not hasattr(aq_base(self.old), 'propertyIds'):
            # no properties available
            return None

        for id in self.old.propertyIds():
            if id in self._properties_ignored:
                # migrated by dc or other
                continue
            self._checkLoadAttr(id)

            value = self.old.getProperty(id)
            typ = self.old.getPropertyType(id)

            self._properties += ({'id': id,
                                  'type': typ,
                                  'value': value},)

    def migrate_properties(self):
        """Migrates zope properties

        Removes the old (if exists) and adds a new
        """
        if not hasattr(aq_base(self.new), '_delProperty'):
            # no properties available
            return None

        for prop_d in self._properties:
            id = prop_d['id']
            value = prop_d['value']
            typ = prop_d['type']
            __traceback_info__ = (self.new, id, value, typ)
            if self.new.hasProperty(id):
                self.new._delProperty(id)

            # continue if the object already has this attribute
            if getattr(aq_base(self.new), id, _marker) is not _marker:
                continue
            try:
                self.new.manage_addProperty(id, value, typ)
            except ConflictError:
                raise
            except:
                LOG.error('Failed to set property %s type %s to %s '
                          'at object %s' % (id, typ, value, self.new),
                          exc_info=True)

    def beforeChange_owner(self):
        """Load the zope owner."""
        self._checkLoadAttr('owner')
        self._checkLoadAttr('_owner')

        # getWrappedOwner is not always available
        if hasattr(aq_base(self.old), 'getWrappedOwner'):
            self.owner = self.old.getWrappedOwner()
        else:
            # fallback: not very nice but at least it works trying to
            # get/set the owner via getOwner(), changeOwnership(...)
            # did not work, at least not with plone 1.x, at 1.0.1,
            # zope 2.6.2
            self._owner = self.old.getOwner(info = 1)

    def migrate_owner(self):
        """Migrates the zope owner."""
        if hasattr(self, 'owner'):
            self.new.changeOwnership(self.owner)
        else:
            self.new._owner = self._owner

    def beforeChange_localroles(self):
        """Load local roles."""
        self._checkLoadAttr('__ac_local_roles__')
        self._checkLoadAttr('__ac_local_roles_block__')
        self.__ac_local_roles__ = self.old.__ac_local_roles__
        if hasattr(self.old, '__ac_local_roles_block__'):
            self.__ac_local_roles_block__ = (
                self.old.__ac_local_roles_block__)

    def migrate_localroles(self):
        """Migrate local roles
        """
        self.new.__ac_local_roles__ = None
        # Clean the auto-generated creators by Archetypes
        # ExtensibleMetadata.  Do set the original creators.
        self.new.setCreators(self.old.listCreators())
        if not self.__ac_local_roles__:
            # Only set owner local role if the owner can be retrieved
            if self.owner is not SpecialUsers.nobody:
                self.new.manage_setLocalRoles(
                    self.owner.getId(), ['Owner'])
        else:
            self.new.__ac_local_roles__ = copy(
                self.__ac_local_roles__)
            if hasattr(self, '__ac_local_roles_block__'):
                self.new.__ac_local_roles_block__ = (
                    self.__ac_local_roles_block__)

    def beforeChange_permission_settings(self):
        """Load permission settings (permission <-> role)."""
        self._checkLoadAttr('ac_inherited_permissions')
        self.ac_inherited_permissions = getPermissionMapping(
            self.old.ac_inherited_permissions(1))

    def migrate_permission_settings(self):
        """Migrate permission settings (permission <-> role)

        The acquire flag is coded into the type of the sequence. If roles is a list
        than the roles are also acquire. If roles is a tuple the roles aren't
        acquired.
        """
        _marker = []

        oldmap = self.ac_inherited_permissions
        newmap = getPermissionMapping(self.new.ac_inherited_permissions(1))
        for key, values in oldmap.items():
            old_p = Permission(key, values, self.old)
            old_roles = old_p.getRoles(default=_marker)
            if old_roles is _marker:
                continue
            new_values = newmap.get(key, ())
            new_p = Permission(key, new_values, self.new)
            new_p.setRoles(old_roles)

    def beforeChange_withmap(self):
        """Loads other attributes from obj.__dict__.  See migrat_withmap."""

        for oldKey, newKey in self.map.items():
            if not newKey:
                newKey = oldKey
                self._checkLoadAttr(newKey)

            oldVal = getattr(self.old, oldKey)
            if callable(oldVal):
                value = oldVal()
            else:
                value = oldVal

            setattr(self, newKey, value)

    def migrate_withmap(self):
        """Migrates other attributes from obj.__dict__ using a map

        The map can contain both attribute names and method names

        'oldattr' : 'newattr'
            new.newattr = oldattr
        'oldattr' : ''
            new.oldattr = oldattr
        'oldmethod' : 'newattr'
            new.newattr = oldmethod()
        'oldattr' : 'newmethod'
            new.newmethod(oldatt)
        'oldmethod' : 'newmethod'
            new.newmethod(oldmethod())"""
        for newKey in self.map.values():
            newVal = getattr(self.new, newKey)
            value = getattr(self, newKey)
            if callable(newVal):
                newVal(value)
            else:
                setattr(self.new, newKey, value)

class BaseInplaceCMFMigrator(BaseInplaceMigrator, BaseCMFMigrator):
    """Support for CMF migration that doesn't require renaming the old
    object."""

    def beforeChange_dc(self):
        """Load dublin core metadata."""
        for accessor, mutator in METADATA_MAPPING:
            self._checkLoadAttr(mutator)

            oldAcc = getattr(self.old, accessor)
            oldValue = oldAcc()
            setattr(self, mutator, oldValue)

    def migrate_dc(self):
        """Migrates dublin core metadata This needs to be done after
        custom migrations, as you cannot setContentType on Images and
        Files until there is an object stored.  """
        for accessor, mutator in METADATA_MAPPING:
            newMut = getattr(self.new, mutator)
            oldValue = getattr(self, mutator)
            newMut(oldValue)

    def beforeChange_workflow(self):
        """Load the workflow state."""
        self._checkLoadAttr('workflow_history')
        self.workflow_history = getattr(self.old, 'workflow_history',
                                        None)

    def migrate_workflow(self):
        """migrate the workflow state."""
        wfh = self.workflow_history
        if wfh:
            wfh = copyPermMap(wfh)
            self.new.workflow_history = wfh

    def beforeChange_allowDiscussion(self):
        """load allow discussion bit."""
        self._checkLoadAttr('isDiscussable')
        if (getattr(aq_base(self.old), 'allowDiscussion', _marker) is
            not _marker):
            self.isDiscussable = self.old.allowDiscussion()

    def migrate_allowDiscussion(self):
        """migrate allow discussion bit."""
        if (getattr(aq_base(self.new), 'isDiscussable', _marker)
            is not _marker and hasattr(self, 'isDiscussable')):
            self.new.isDiscussable(self.isDiscussable)

    def beforeChange_discussion(self):
        """Load talkback discussion bit."""
        self._checkLoadAttr('talkback')
        self.talkback = getattr(self.old.aq_inner.aq_explicit,
                                'talkback', _marker)

    def migrate_discussion(self):
        """migrate talkback discussion bit
        """
        if self.talkback is _marker:
            return
        else:
            self.new.talkback = self.talkback

class InplaceItemMigrationMixin(ItemMigrationMixin):
    """Migrates a non folderish object."""

    def renameOld(self):
        """Remove the old object rather than rename it."""
        self.parent.manage_delObjects([self.orig_id])

    def remove(self):
        """Since we remove where other migrators rename, we pass
        here."""
        pass

class InplaceFolderMigrationMixin(InplaceItemMigrationMixin,
                                  FolderMigrationMixin):
    """Migrates a folderish object inplace."""
    isFolderish = True

class InplaceUIDMigrator(object):
    """Inplace Migrator class for migration CMF and AT uids. """

    def beforeChange_cmf_uid(self):
        """Load CMF uids."""
        self._checkLoadAttr('uid')
        uidhandler = getToolByName(self.parent, 'portal_uidhandler',
                                   None)
        if uidhandler is not None:
            self.uid = uidhandler.queryUid(self.old, default=None)

    def migrate_cmf_uid(self):
        """Migrate CMF uids."""
        uidhandler = getToolByName(self.parent, 'portal_uidhandler', None)
        if uidhandler is not None:
            return # no uid handler available
        uid = self.uid
        if uid is not None:
            uidhandler.setUid(self.new, uid, check_uniqueness=False)

    def beforeChange_at_uuid(self):
        """Load AT universal uid."""
        self._checkLoadAttr('UID')
        if IReferenceable.providedBy(self.old):
            self.UID = self.old.UID()
            self.old._uncatalogUID(self.parent)
        else:
            self.UID = None

    def migrate_at_uuid(self):
        """Migrate AT universal uid
        """
        uid = self.UID
        if uid:
            self.new._setUID(uid)

class InplaceCMFItemMigrator(InplaceItemMigrationMixin,
                             InplaceUIDMigrator,
                             BaseInplaceCMFMigrator):
    """Inplace migrator for items implementing the CMF API."""
    pass

class InplaceCMFFolderMigrator(InplaceFolderMigrationMixin,
                               InplaceUIDMigrator,
                               BaseInplaceCMFMigrator):
    """Inplace migrator from folderish items implementing the CMF
    API."""
    pass



"""Migration tools for ATContentTypes

Migration system for the migration from CMFDefault/Event types to archetypes
based CMFPloneTypes (http://plone.org/products/atcontenttypes/).

Copyright (c) 2004-2005, Christian Heimes <tiran@cheimes.de> and contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.
"""
__author__  = 'Christian Heimes <tiran@cheimes.de>'
__docformat__ = 'restructuredtext'

# Catalog patches for ATCT migration
#
# The catalog patches are patching catalog_object and uncatalog_object in order to
# speed up the migration. Especially for folder migration with hundreds or more
# children the speed gain can be immense.
#
# The patch is altering uncatalog_object to do nothing so that objects aren't
# uncataloged. catalog_object is altered to update only a small subset of the
# indexes and no metadata. Only indexes that are used in the migration are updated.
#
# applyCatalogPatch and removeCatalogPatch must be wrapped in a try/finally block!
#
# indexObject(), unindexObject() and reindexObject() are using catalog_object() and
# uncatalog_object(). There is no need to patch these methods.

import logging
# instancemethod(function, instance, class)
# Create an instance method object.
from types import MethodType as instancemethod

from Products.CMFCore.utils import getToolByName

LOG = logging.getLogger('ATCT.migration')
UPDATE_METADATA = False
IDXS = ('meta_type', 'portal_type', 'path')

def uncatalog_object(self, uid):
    """NOOP uncatalog_object
    """
    pass

def catalog_object(self, obj, uid=None, idxs=None, update_metadata=1,
                   pghandler=None):
   """Minimalistic catalog_object method
   """
   return self._atct_catalog_object(obj, uid=uid, idxs=IDXS,
                                    update_metadata=UPDATE_METADATA,
                                    pghandler=None)

def applyCatalogPatch(portal):
    """Patch catalog
    """
    catalog = getToolByName(portal, 'portal_catalog')
    klass = catalog.__class__
    LOG.info('Patching catalog_object and uncatalog_object of %s' %
             catalog.absolute_url(1))
    if hasattr(klass, '_atct_catalog_object'):
        raise RuntimeError("%s already has _atct_catalog_object" % catalog)
    if hasattr(klass, '_atct_uncatalog_object'):
        raise RuntimeError("%s already has _atct_uncatalog_object" % catalog)

    klass._atct_catalog_object = klass.catalog_object.im_func
    klass.catalog_object = instancemethod(catalog_object, None, klass)

    klass._atct_uncatalog_object = klass.uncatalog_object.im_func
    klass.uncatalog_object = instancemethod(uncatalog_object, None, klass)

    return klass

def removeCatalogPatch(klass):
    """Unpatch catalog

    removeCatalogPatch must be called with the catalog class and not with the portal
    as argument. If migration fails the transaction is aborted explictly. The portal
    doesn't exist any longer and the method would be unable to acquire the
    portal_catalog.
    """
    LOG.info('Unpatching catalog_object and uncatalog_object')

    if hasattr(klass, '_atct_catalog_object'):
        klass.catalog_object = klass._atct_catalog_object.im_func
        del klass._atct_catalog_object

    if hasattr(klass, '_atct_uncatalog_object'):
        klass.uncatalog_object = klass._atct_uncatalog_object.im_func
        del klass._atct_uncatalog_object



from Products.contentmigration.migrator import InlineFieldActionMigrator
from Products.contentmigration.walker import CustomQueryWalker
from Products.contentmigration.tests.layer import TestLayer
from Products.contentmigration.tests.layer import SchemaExtenderTestLayer

import transaction

from plone.app.testing.bbb_at import PloneTestCase

# Callback methods

def makeUpper(obj, val, **kwargs):
    s = str(val)
    s = s.upper()
    val.update(s, obj)
    return val

def conditionallyAbortObject(obj, **kwargs):
    if obj.getId() == 'd1':
        return False
    else:
        return True

def conditionallyAbortAttribute(obj, attributeName, attributeValue, **kwargs):
    if 'd1' in obj.getId():
        return False
    else:
        return True

def callAfterObject(obj, **kwargs):
    kwargs['lst'].append(obj.getId())

def callAfterAttribute(obj, attributeName, attributeValue, **kwargs):
    kwargs['lst'].append("%s: %s = %s" % (obj.getId(), attributeName, attributeValue,))

# Test migrator
class TestMigrator(InlineFieldActionMigrator):
    src_portal_type = 'Document'
    src_meta_type = 'ATDocument'

class ContentMigratorTestCase(PloneTestCase):

    layer = TestLayer

    class Session(dict):
        def set(self, key, value):
            self[key] = value

    def _setup(self):
        PloneTestCase.PloneTestCase._setup(self)
        self.app.REQUEST['SESSION'] = self.Session()

    def execute(self, query, actions, callBefore=None, **kwargs):
        TestMigrator.fieldActions = actions
        walker = CustomQueryWalker(self.portal, TestMigrator,
                                    query = query,
                                    callBefore = callBefore,
                                    **kwargs)
        # Need this to avoid copy errors....
        transaction.savepoint(optimistic=True)
        walker.go(**kwargs)


class SchemaExtenderMigratorTestCase(ContentMigratorTestCase):

    layer = SchemaExtenderTestLayer



from App.class_init import InitializeClass
from Products.CMFCore.PortalContent import PortalContent
from zope.component.factory import Factory

def addDocument(self, id, title='', description='', text_format='', text=''):
    """Add a Document.
    """
    o = Document(id, title, description, text_format, text)
    self._setObject(id, o, suppress_events=True)


class Document(PortalContent):
    """ A Document """

    def __init__(self, id, title='', description='', text_format='', text=''):
        self.id = id
        self.title = title
        self.description = description

    def Title(self):
        return self.title

    def listCreators(self):
        return ('test_user_1_',)

    def Creator(self):
        return self.listCreators()[0]


InitializeClass(Document)
DocumentFactory = Factory(Document)



"""Walkers for Migration suite

Migration system for the migration from CMFDefault/Event types to archetypes
based CMFPloneTypes (http://plone.org/products/atcontenttypes/).

Copyright (c) 2004-2005, Christian Heimes <tiran@cheimes.de> and contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.
"""
from __future__ import print_function
__author__  = 'Christian Heimes <tiran@cheimes.de>'
__docformat__ = 'restructuredtext'

from Products.CMFCore.utils import getToolByName
from Products.contentmigration.common import HAS_LINGUA_PLONE
from Products.contentmigration.common import registerWalker
from six.moves import cStringIO as StringIO
from ZODB.POSException import ConflictError

import logging
import traceback
import transaction


LOG = logging.getLogger('ATCT.migration')

# archetypes.schemaextender addon?
try:
    # beware of moving this, the top-level package has an archetypes module
    from archetypes.schemaextender.extender import disableCache
    disableCache
except ImportError:
    def disableCache(request):
        pass


class StopWalking(StopIteration):
    pass


class MigrationError(RuntimeError):
    def __init__(self, path, migrator, traceback):
        self.src_portal_type = migrator.src_portal_type
        self.dst_portal_type = migrator.dst_portal_type
        self.tb = traceback
        self.path = path

    def __str__(self):
        return "MigrationError for obj at %s (%s -> %s):\n%s" % (self.path,
                    self.src_portal_type, self.dst_portal_type, self.tb)

class Walker(object):
    """Walks through the system and migrates every object it finds

    arguments:
    * portal
      portal root object as context
    * migrator
      migrator class
    * src and dst_portal_type
      ids of the portal types to migrate
    * transaction_size (int)
      Amount of objects before a transaction, subtransaction or new savepoint is
      created. A small number might slow down the process since transactions are
      possible costly.
    * full_transaction
      Commit a full transaction after transaction size
    * use_savepoint
      Create savepoints and roll back to the savepoint if an error occurs
    * limit
      Limits the catalog query to at most x items

    full_transaction and use_savepoint are mutual exclusive.
    o When the default values (both False) are used a subtransaction is committed.
      If an error occurs *all* changes are lost.
    o If full_transaction is enabled a full transaction is committed. If an error
      occurs the migration process stops and all changes sine the last transaction
      are lost.
    o If use_savepoint is set savepoints are used. A savepoint is like a
      subtransaction which can be rolled back. If an errors occurs the transaction
      is rolled back to the last savepoint and the migration goes on. Some objects
      will be left unmigrated.

    """

    def __init__(self, portal, migrator, src_portal_type=None, dst_portal_type=None,
                 **kwargs):
        self.portal = portal
        self.catalog = getToolByName(portal, 'portal_catalog')
        self.migrator = migrator
        if src_portal_type is None:
            self.src_portal_type = self.migrator.src_portal_type
        else:
            self.src_portal_type = src_portal_type
        if dst_portal_type is None:
            self.dst_portal_type = self.migrator.dst_portal_type
        else:
            self.dst_portal_type = dst_portal_type
        self.src_meta_type = self.migrator.src_meta_type
        self.dst_meta_type = self.migrator.dst_meta_type

        self.transaction_size = int(kwargs.get('transaction_size', 20))
        self.full_transaction = kwargs.get('full_transaction', False)
        self.use_savepoint = kwargs.get('use_savepoint', False)
        self.limit = kwargs.get('limit', False)

        if self.full_transaction and self.use_savepoint:
            raise ValueError

        self.out = StringIO()
        self.counter = 0
        self.errors = []

        # Disable schema cache
        request = getattr(portal, 'REQUEST', None)
        if request is not None:
            disableCache(request)

    def go(self, **kwargs):
        """runner

        Call it to start the migration
        """
        # catalog subtransaction conflict w/ savepoints because a subtransaction
        # destroys all existing savepoints.
        # disable subtransactions for all known catalogs and restore them later
        old_thresholds = {}
        for id in ('portal_catalog', 'uid_catalog', 'reference_catalog'):
            catalog = getToolByName(self.portal, id, None)
            if catalog is not None:
                old_thresholds[id] = getattr(self.catalog, 'threshold', None)
                catalog.threshold = None

        try:
            self.migrate(self.walk(), **kwargs)
        finally:
            for id, threshold in old_thresholds.items():
                catalog = getToolByName(self.portal, id, None)
                if catalog is not None:
                    catalog.threshold = threshold

    __call__ = go

    def walk(self):
        """Walks around and returns all objects which needs migration

        :return: objects (with acquisition wrapper) that needs migration
        :rtype: list of objects
        """
        raise NotImplementedError

    def migrate(self, objs, **kwargs):
        """Migrates the objects in the ist objs
        """
        out = self.out
        counter = self.counter
        errors = self.errors
        full_transaction = self.full_transaction
        transaction_size = self.transaction_size
        use_savepoint = self.use_savepoint

        src_portal_type = self.src_portal_type
        dst_portal_type = self.dst_portal_type

        if use_savepoint:
            savepoint = transaction.savepoint()

        for obj in objs:
            objpath = '/'.join(obj.getPhysicalPath())
            msg = 'Migrating %s (%s -> %s)' % (objpath, src_portal_type,
                                               dst_portal_type)
            LOG.debug(msg)
            print(msg, file=out)
            counter+=1

            migrator = self.migrator(obj,
                                     src_portal_type=src_portal_type,
                                     dst_portal_type=dst_portal_type,
                                     **kwargs)

            try:
                # run the migration
                migrator.migrate()
            except ConflictError:
                raise
            except: # except all!
                msg = "Failed migration for object %s (%s -> %s)" %  (objpath,
                           src_portal_type, dst_portal_type)
                # printing exception
                f = StringIO()
                traceback.print_exc(limit=None, file=f)
                tb = f.getvalue()

                LOG.error(msg, exc_info = True)
                errors.append({'msg' : msg, 'tb' : tb, 'counter': counter})

                if use_savepoint:
                    if savepoint.valid:
                        # Rollback to savepoint
                        LOG.info("Rolling back to last safe point")
                        print(msg, file=out)
                        print(tb, file=out)
                        savepoint.rollback()
                        # XXX: savepoints are invalidated once they are used
                        savepoint = transaction.savepoint()
                        continue
                    else:
                        LOG.error("Savepoint is invalid. Probably a subtransaction "
                            "was committed. Unable to roll back!")
                #  stop migration process after an error
                # aborting transaction
                LOG.error("FATAL: Migration has failed, aborting transaction!")
                transaction.abort()
                raise MigrationError(objpath, migrator, tb)

            if counter % transaction_size == 0:
                if full_transaction:
                    transaction.commit()
                    LOG.debug('Transaction comitted after %s objects' % counter)
                elif use_savepoint:
                    LOG.debug('Creating new safepoint after %s objects' % counter)
                    savepoint = transaction.savepoint()
                else:
                    LOG.debug('Committing subtransaction after %s objects' % counter)
                    transaction.savepoint(optimistic=True)

        self.out = out
        self.counter = counter
        self.errors = errors

    def getOutput(self):
        """Get migration notes

        :return: objects (with acquisition wrapper) that needs migration
        :rtype: list of objects
        """
        return self.out.getvalue()

class CatalogWalker(Walker):
    """Walker using portal_catalog
    """

    def walk(self):
        """Walks around and returns all objects which needs migration

        :return: objects (with acquisition wrapper) that needs migration
        :rtype: generator
        """
        catalog = self.catalog
        query = {
            'portal_type' : self.src_portal_type,
            'meta_type' : self.src_meta_type,
            'path' : "/".join(self.portal.getPhysicalPath()),
        }
        if HAS_LINGUA_PLONE and 'Language' in catalog.indexes():
            query['Language'] = 'all'

        brains = catalog(query)
        limit = getattr(self, 'limit', False)
        if limit:
            brains = brains[:limit]

        # unpack to be able to log brains that break during migration
        brains = [i for i in brains]

        for brain in brains:
            try:
                obj = brain.getObject()
            except AttributeError:
                LOG.error("Couldn't access %s" % brain.getPath())
                continue
            except KeyError:
                LOG.error("Couldn't access RID %s" % brain.getRID())
                continue
            try:
                state = obj._p_changed
            except:
                state = 0
            if obj is not None:
                yield obj
                # safe my butt
                if state is None:
                    obj._p_deactivate()

registerWalker(CatalogWalker)

class CatalogWalkerWithLevel(Walker):
    """Walker using the catalog but only returning objects for a specific depth

    Requires ExtendedPathIndex!
    """

    def __init__(self, portal, migrator, src_portal_type=None, dst_portal_type=None,
                 depth=1, max_depth=100, **kwargs):
        Walker.__init__(self, portal, migrator, src_portal_type, dst_portal_type,
                        **kwargs)
        self.depth=depth
        self.max_depth = max_depth

    def walk(self):
        """Walks around and returns all objects which needs migration

        :return: objects (with acquisition wrapper) that needs migration
        :rtype: generator

        TODO: stop when no objects are left. Don't try to migrate until the walker
              reaches max_depth
        """
        depth = self.depth
        max_depth = self.max_depth
        catalog = self.catalog
        root = '/'.join(self.portal.getPhysicalPath())
        rootlen = len(root)
        query = {
            'portal_type' : self.src_portal_type,
            'meta_type' : self.src_meta_type,
            'path' : {'query' : root, 'depth' : depth},
        }

        if HAS_LINGUA_PLONE and 'Language' in catalog.indexes():
            #query['Language'] = catalog.uniqueValuesFor('Language')
            query['Language'] = 'all'

        while True:
            if depth > max_depth:
                raise StopWalking
            query['path']['depth'] = depth
            for brain in catalog(query):
                # depth 'n' returns objects with depth of 'n' and *smaller*
                # but we want to migrate only object with a depth of
                # exactly 'n'
                relpath = brain.getPath()[rootlen:]
                if not relpath.count('/') == depth:
                    continue

                obj = brain.getObject()
                try: state = obj._p_changed
                except: state = 0
                if obj is not None:
                    yield obj
                    # safe my butt
                    if state is None: obj._p_deactivate()

            depth+=1

registerWalker(CatalogWalkerWithLevel)

def useLevelWalker(portal, migrator, **kwargs):
    w = CatalogWalkerWithLevel(portal, migrator)
    return w.go(**kwargs)



# -*- coding: utf-8 -*-
"""Migration tools for ATContentTypes

Migration system for the migration from CMFDefault/Event types to archetypes
based CMFPloneTypes (http://plone.org/products/atcontenttypes/).

Copyright (c) 2004, Christian Heimes <tiran@cheimes.de> and contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.


"""
from __future__ import print_function
__author__  = 'Christian Heimes <tiran@cheimes.de>'
__docformat__ = 'restructuredtext'

from Products.contentmigration.catalogpatch import applyCatalogPatch
from Products.contentmigration.catalogpatch import removeCatalogPatch
from Products.CMFCore.utils import getToolByName

import logging
import pkg_resources
import sys


# Is there a multilingual addon?
try:
    pkg_resources.get_distribution('Products.LinguaPlone')
except pkg_resources.DistributionNotFound:
    HAS_LINGUA_PLONE = False
else:
    HAS_LINGUA_PLONE = True


LOG = logging.getLogger('ATCT.migration')

# This method was coded by me (Tiran) for CMFPlone. I'm maintaining a copy here
# to avoid dependencies on CMFPlone
def _createObjectByType(type_name, container, id, *args, **kw):
    """Create an object without performing security checks

    invokeFactory and fti.constructInstance perform some security checks
    before creating the object. Use this function instead if you need to
    skip these checks.

    This method uses some code from
    CMFCore.TypesTool.FactoryTypeInformation.constructInstance
    to create the object without security checks.

    It doesn't finish the construction and so doesn't reinitializes the workflow.
    """
    id = str(id)
    typesTool = getToolByName(container, 'portal_types')
    fti = typesTool.getTypeInfo(type_name)
    if not fti:
        raise ValueError('Invalid type %s' % type_name)

    ob = fti._constructInstance(container, id, *args, **kw)

    # in CMF <2.2 the portal type hasn't been set yet
    if 'portal_type' not in ob.__dict__ and hasattr(ob, '_setPortalTypeName'):
        ob._setPortalTypeName(fti.getId())

    return ob

from Acquisition import aq_base
from App.Dialogs import MessageDialog
#from OFS.CopySupport import CopyContainer
from OFS.CopySupport import CopyError
from cgi import escape

from six.moves import cStringIO as StringIO


def unrestricted_rename(self, id, new_id):
    """Rename a particular sub-object

    Copied from OFS.CopySupport

    Less strict version of manage_renameObject:
        * no write look check
        * no verify object check from PortalFolder so it's allowed to rename
          even unallowed portal types inside a folder
    """
    try: self._checkId(new_id)
    except: raise CopyError(MessageDialog(
                  title='Invalid Id',
                  message=sys.exc_info()[1],
                  action ='manage_main'))
    ob=self._getOb(id)
    #!#if ob.wl_isLocked():
    #!#    raise(ResourceLockedError, 'Object "%s" is locked via WebDAV' % ob.getId())
    if not ob.cb_isMoveable():
        raise CopyError('Not supported {}'.format(escape(id)))
    #!#self._verifyObjectPaste(ob)
    #!#CopyContainer._verifyObjectPaste(self, ob)
    try:    ob._notifyOfCopyTo(self, op=1)
    except: raise(CopyError, MessageDialog(
                  title='Rename Error',
                  message=sys.exc_info()[1],
                  action ='manage_main'))
    self._delObject(id)
    ob = aq_base(ob)
    ob._setId(new_id)

    # Note - because a rename always keeps the same context, we
    # can just leave the ownership info unchanged.
    self._setObject(new_id, ob, set_owner=0)
    ob = self._getOb(new_id)
    ob._postCopy(self, op=1)

    #!#if REQUEST is not None:
    #!#    return self.manage_main(self, REQUEST, update_menu=1)
    return None

class Registry(dict):
    """Common registry
    """

    def register(self, cls):
        self[cls.__name__] = cls

class MigratorRegistry(Registry):
    """Migrator Registry
    """

    def registerATCT(self, cls, for_cls):
        """Special register method for ATCT based migrators
        """
        cls.src_portal_type = for_cls._atct_newTypeFor['portal_type']
        cls.src_meta_type = for_cls._atct_newTypeFor['meta_type']
        cls.dst_portal_type = for_cls.portal_type
        cls.dst_meta_type = for_cls.meta_type
        self.register(cls)

    def register(self, cls):
        key = (cls.src_meta_type, cls.dst_meta_type)
        assert key not in self
        self[key] = cls
        self[cls.__name__] = cls

class WalkerRegistry(Registry):
    """Walker Registry
    """
    pass

_migratorRegistry = MigratorRegistry()
registerMigrator = _migratorRegistry.register
registerATCTMigrator = _migratorRegistry.registerATCT
listMigrators = _migratorRegistry.items
getMigrator = _migratorRegistry.get

_walkerRegistry = WalkerRegistry()
registerWalker = _walkerRegistry.register
listWalkers = _walkerRegistry.items

def migratePortalType(portal, src_portal_type, dst_portal_type, out=None,
                      migrator=None, use_catalog_patch=False, **kwargs):
    """Migrate from src portal type to dst portal type

    Additional **kwargs are applied to the walker
    """
    if not out:
        out = StringIO()

    # migrators are also registered by (src meta type, dst meta type)
    # let's find the right migrator for us
    ttool = getToolByName(portal, 'portal_types')
    src = ttool.getTypeInfo(src_portal_type)
    dst = ttool.getTypeInfo(dst_portal_type)
    if src is None or dst is None:
        raise ValueError('Unknown src or dst portal type: %s -> %s' % (
                           src_portal_type, dst_portal_type,))

    key = (src.Metatype(), dst.Metatype())
    migratorFromRegistry = getMigrator(key)
    if migratorFromRegistry is None:
        raise ValueError('No registered migrator for "%s" found' % str(key))

    if migrator is not None:
        # got a migrator, make sure it is the right one
        if migrator is not migratorFromRegistry:
            raise ValueError('ups')
    else:
        migrator = migratorFromRegistry

    Walker = migrator.walkerClass

    msg = '--> Migrating %s to %s with %s' % (src_portal_type,
           dst_portal_type, Walker.__name__)
    if use_catalog_patch:
        msg+=', using catalog patch'
    if kwargs.get('use_savepoint', False):
        msg+=', using savepoints'
    if kwargs.get('full_transaction', False):
        msg+=', using full transactions'

    print(msg, file=sys.stderr)
    LOG.debug(msg)

    walk = Walker(portal, migrator, src_portal_type=src_portal_type,
                  dst_portal_type=dst_portal_type, **kwargs)
    # wrap catalog patch inside a try/finally clause to make sure that the catalog
    # is unpatched under *any* circumstances (hopely)
    try:
        if use_catalog_patch:
            catalog_class = applyCatalogPatch(portal)
        walk.go()
    finally:
        if use_catalog_patch:
            removeCatalogPatch(catalog_class)

    print(walk.getOutput(), file=out)
    LOG.debug('<-- Migrating %s to %s done' % (src_portal_type, dst_portal_type))

    return out






from setuptools import setup, find_packages

import sys


if sys.version_info[0] != 2:
    # Prevent creating or installing a distribution with Python 3.
    raise ValueError("Products.contentmigration is based on Archetypes, which is Python 2 only.")

version = '2.2.2'

setup(
    name='Products.contentmigration',
    version=version,
    description="A generic content migration framework for Plone.",
    long_description=(open("README.rst").read() + "\n" +
                      open("CHANGES.rst").read()),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 5.2",
        "Framework :: Plone :: Core",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
    ],
    keywords='Plone Archetypes ATContentTypes',
    author='Martin Aspeli (and others)',
    author_email='plone-developers@lists.sourceforge.net',
    url='https://pypi.org/project/Products.contentmigration',
    license='LGPL',
    packages=find_packages(),
    namespace_packages=['Products'],
    include_package_data=True,
    zip_safe=False,
    python_requires='==2.7.*',
    extras_require=dict(
        test=[
            'archetypes.schemaextender',
            'plone.app.testing',
            'Products.CMFPlone',
            'six',
            'zope.testing',
            'Zope2',
        ],
    ),
    install_requires=[
        'setuptools',
    ],
)



from plone.app.testing import bbb_at
from plone.app import testing
from plone.testing import z2
from Products.GenericSetup import EXTENSION, profile_registry


def setupSampleTypeProfile():
    profile_registry.registerProfile('CMF_sampletypes',
        'CMF Sample Content Types',
        'Extension profile including CMF sample content types',
        'profiles/testing',
        'Products.contentmigration',
        EXTENSION)


class PloneTestCaseFixture(bbb_at.PloneTestCaseFixture):

    defaultBases = (bbb_at.PTC_FIXTURE, )

    def setUpZope(self, app, configurationContext):
        setupSampleTypeProfile()
        from Products import contentmigration
        self.loadZCML('testing.zcml', package=contentmigration)
        self.loadZCML('testing-schemaextender.zcml', package=contentmigration)
        z2.installProduct(app, 'Products.contentmigration')

    def setUpPloneSite(self, portal):
        testing.applyProfile(portal, 'Products.contentmigration:CMF_sampletypes')


PCM_FIXTURE = PloneTestCaseFixture()
TestLayer = testing.FunctionalTesting(
    bases=(PCM_FIXTURE, ), name='PloneContentMigrationTestCase:Functional')

SchemaExtenderTestLayer = TestLayer
